//
// Created by Tom on 07/03/2023.
//

#ifndef TYPON_BUILTINS_HPP
#define TYPON_BUILTINS_HPP

#include <fmt/format.h>
#include <iostream>
#include <optional>
#include <ostream>
#include <string>

#include <python/basedef.hpp>
#include <typon/typon.hpp>

#ifdef __cpp_lib_unreachable
#include <utility>
[[noreturn]] inline void TYPON_UNREACHABLE() { std::unreachable(); }
#else
#include <cstdlib>
[[noreturn]] inline void TYPON_UNREACHABLE() { std::abort(); }
#endif

#define _Args(...) __VA_ARGS__
#define COMMA() ,
#define METHOD(ret, name, args, ...)                                           \
  static constexpr struct name##_s : method   {                                                          \
    template <typename Self> ret operator() args const __VA_ARGS__             \
  } name{};

#define METHOD_GEN(gen, ret, name, args, ...)                                           \
  static constexpr struct name##_s : method   {                                                          \
    template <typename Self, _Args gen> ret operator() args const __VA_ARGS__             \
  } name{};

#define FUNCTION(ret, name, args, ...)                                         \
  struct {                                                                     \
    ret operator() args __VA_ARGS__                                            \
  } static constexpr name;

using namespace std::literals;

template <typename T>
concept PySmartPtr = requires { typename T::element_type; };

template <typename T>
concept PyUserType = requires { typename T::py_type; };

template <typename T> struct RealType {
  using type = T;
};

template <PyUserType T> struct RealType<T> {
  using type = typename T::py_type;
};

template <PySmartPtr T> struct RealType<T> {
  using type = typename T::element_type;
};

template <typename T> using PyObj = std::shared_ptr<typename RealType<T>::type>;

template <typename T, typename... Args> auto pyobj(Args &&...args) -> PyObj<T> {
  return std::make_shared<typename RealType<T>::type>(
      std::forward<Args>(args)...);
}

template <typename T, typename... Args> auto pyobj_agg(Args &&...args) -> PyObj<T> {
  return std::make_shared<typename RealType<T>::type>((typename RealType<T>::type) { std::forward<Args>(args)... });
}

// typon_len

template <typename T>
concept PyIterator = requires(T t) {
  { t.py_next() } -> std::same_as<std::optional<T>>;
};

template <typename T>
concept PyIterable = requires(T t) {
  { t.py_iter() } -> PyIterator;
};

template <PyIterable T, PyIterator U> U iter(const T &t) { return t.py_iter(); }

template <typename T>
concept CppSize = requires(const T &t) {
  { t.size() } -> std::same_as<size_t>;
};

template <typename T>
concept PyLen = requires(const T &t) {
  { t.py_len() } -> std::same_as<size_t>;
};

template <CppSize T>
  requires(!PyLen<T>)
size_t len(const T &t) {
  return t.size();
}

template <PyLen T> size_t len(const T &t) { return t.py_len(); }

template <typename T>
concept PyNext = requires(T t) {
  { t.py_next() } -> std::same_as<std::optional<typename T::value_type>>;
};

struct {
  template <PyNext T>
  std::optional<typename T::value_type>
  operator()(T &t, std::optional<typename T::value_type> def = std::nullopt) {
    auto opt = t.py_next();
    return opt ? opt : def;
  }
} next;

template <typename T>
std::ostream &operator<<(std::ostream &os, std::optional<T> const &opt) {
  return opt ? os << opt.value() : os << "None";
}

bool is_cpp() { return true; }

class NoneType {
public:
  template <typename T> operator T *() const { return nullptr; }

  template <typename T> operator std::optional<T>() const {
    return std::nullopt;
  }
} PyNone{};

#define system_error(err, message)                                             \
  do {                                                                         \
    puts(message);                                                             \
    throw fmt::system_error(err, message);                                     \
  } while (0)

#include "builtins/bool.hpp"
#include "builtins/complex.hpp"
#include "builtins/dict.hpp"
#include "builtins/list.hpp"
#include "builtins/print.hpp"
#include "builtins/range.hpp"
#include "builtins/set.hpp"
#include "builtins/slice.hpp"
#include "builtins/str.hpp"

struct file_s {
  struct py_type {
    METHOD(
        typon::Task<PyStr>, read, (Self self, size_t size = -1), {
          if (size == -1) {
            size = self->len;
          }
          PyStr buf(size, '\0');
          int nbytes = co_await typon::io::read(self->fd, buf.data(), size);
          if (nbytes < 0) {
            system_error(-nbytes, "read()");
          }
          buf.resize(nbytes);
          co_return std::move(buf);
        })
    METHOD(
        typon::Task<int>, write, (Self self, const std::string &buf), {
          int nbytes = co_await typon::io::write(self->fd, buf);
          if (nbytes < 0) {
            system_error(-nbytes, "write()");
          }
          co_return nbytes;
        })
    METHOD(
        typon::Task<void>, close, (Self self),
        { co_await typon::io::close(self->fd); })
    METHOD(
        typon::Task<void>, flush, (Self self),
        { co_await typon::io::fsync(self->fd); })

    py_type(int fd = -1, size_t len = 0) : fd(fd), len(len) {}

    py_type(const py_type &other)
        : fd(other.fd), len(other.len) {}

    METHOD(
        auto, py_enter, (Self self), { return self; })

    METHOD(
        typon::Task<bool>, py_exit, (Self self), {
          co_await dotp(self, close)();
          co_return true;
        })

    int fd;
    size_t len;
  };
} file;

namespace typon {
using PyFile = PyObj<decltype(file)::py_type>;
}

typon::Task<typon::PyFile> open(const PyStr &path, std::string_view mode) {
  const char *path_c = path.c_str();
  size_t len = 0;
  struct statx statxbuf;
  if (int err = co_await typon::io::statx(AT_FDCWD, path_c, 0, STATX_SIZE,
                                          &statxbuf)) {
    // new file
  } else {
    len = statxbuf.stx_size;
  }
  int flags = 0;
  bool created = false, writable = false, readable = false;

  if (mode.find('x') != std::string_view::npos) {
    created = true;
    writable = true;
    flags = O_EXCL | O_CREAT;
  } else if (mode.find('r') != std::string_view::npos) {
    readable = true;
    flags = 0;
  } else if (mode.find('w') != std::string_view::npos) {
    writable = true;
    flags = O_CREAT | O_TRUNC;
  } else if (mode.find('a') != std::string_view::npos) {
    writable = true;
    flags = O_APPEND | O_CREAT;
  }

  if (mode.find('+') != std::string_view::npos) {
    readable = true;
    writable = true;
  }

  if (readable && writable) {
    flags |= O_RDWR;
  } else if (readable) {
    flags |= O_RDONLY;
  } else {
    flags |= O_WRONLY;
  }

  flags |= O_CLOEXEC;

  int fd = co_await typon::io::openat(AT_FDCWD, path_c, flags, 0666);
  if (fd < 0) {
    std::cerr << path << "," << flags << std::endl;
    system_error(-fd, "openat()");
  }
  co_return pyobj<typon::PyFile>(fd, len);
}

#include "../typon/generator.hpp"

#include <pybind11/embed.h>
#include <pybind11/stl.h>
namespace py = pybind11;

#include <utility>

template <typename Ref>
struct lvalue_or_rvalue {

  Ref &&ref;

  template <typename Arg>
  constexpr lvalue_or_rvalue(Arg &&arg) noexcept
      :   ref(std::move(arg))
  { }

  constexpr operator Ref& () const & noexcept { return ref; }
  constexpr operator Ref&& () const && noexcept { return std::move(ref); }
  constexpr Ref& operator*() const noexcept { return ref; }
  constexpr Ref* operator->() const noexcept { return &ref; }

};

namespace typon {
template< class... Types >
using PyTuple = std::tuple<Types...>;


}

template<typename T>
auto& iter_fix_ref(T& obj) { return obj; }

template<PySmartPtr T>
auto& iter_fix_ref(T& obj) { return *obj; }

namespace std {
template <class T> auto begin(std::shared_ptr<T> &obj) { return dotp(obj, begin)(); }

template <class T> auto end(std::shared_ptr<T> &obj) { return dotp(obj, end)(); }
}

template <typename T>
struct AlwaysTrue {                        // (1)
  constexpr bool operator()(const T&) const {
    return true;
  }
};
template <typename Seq>
struct ValueTypeEx {
  using type = decltype(*std::begin(std::declval<Seq&>()));
};
// (2)
template <typename Map, typename Seq, typename Filt = AlwaysTrue<typename ValueTypeEx<Seq>::type>>
auto mapFilter(Map map, Seq seq, Filt filt = Filt()) {

  //typedef typename Seq::value_type value_type;
  using value_type = typename ValueTypeEx<Seq>::type;
  using return_type = decltype(map(std::declval<value_type>()));

  std::vector<return_type> result{};
  for (auto i : seq | std::views::filter(filt)
                    | std::views::transform(map)) result.push_back(i);
  return typon::PyList(std::move(result));
}

#endif // TYPON_BUILTINS_HPP
