from dataclasses import dataclass
from typing import Any, Callable, Optional
from enum import Enum
import operator


@dataclass
class BinOperator:
    symbol: str
    priority: int
    perform: Callable[[float, float], float]


OPERATORS = [
    BinOperator("+", 0, operator.add),
    BinOperator("-", 0, operator.sub),
    BinOperator("*", 1, operator.mul),
    BinOperator("/", 1, operator.truediv)
]

# ops_by_priority = [list(it) for _, it in groupby(OPERATORS, lambda op: op.priority)]
ops_by_priority = [
    [OPERATORS[0], OPERATORS[1]],
    [OPERATORS[2], OPERATORS[3]]
]
MAX_PRIORITY = len(ops_by_priority)
ops_syms = [op.symbol for op in OPERATORS]


class TokenType(Enum):
    NUMBER = 1
    PARENTHESIS = 2
    OPERATION = 3


@dataclass
class Token:
    type: TokenType
    val: str
    num: float


def tokenize(inp: str):
    tokens = []

    index = 0

    def skip_spaces():
        nonlocal index
        while inp[index].isspace():
            index += 1

    def has():
        return index < len(inp)

    def peek():
        return inp[index]

    def read():
        nonlocal index
        index += 1
        return inp[index - 1]

    def read_number():
        res = ""

        while True:
            res += read()
            if not has() or peek() not in "0123456789.":
                break

        return Token(TokenType.NUMBER, res, float(res))  # if "." in res else int(res))

    while has():
        skip_spaces()

        next = peek()

        tok: Token

        if next in ops_syms:
            tok = Token(TokenType.OPERATION, read(), 0)
        elif next in "()":
            tok = Token(TokenType.PARENTHESIS, read(), 0)
        elif next in "0123456789.":
            tok = read_number()
        else:
            raise Exception("invalid character '{}' at {}".format(next, index))

        tokens.append(tok)

    return tokens


def parse(tokens: list[Token]):
    index = 0

    def has():
        return index < len(tokens)

    def current():
        if not has():
            raise Exception("expected token, got EOL")
        return tokens[index]

    def match(type: TokenType, val: Optional[str] = None):
        return has() and tokens[index].type == type and (val is None or tokens[index].val == val)

    def accept(type: TokenType, val: Optional[str] = None):
        nonlocal index
        if match(type, val):
            index += 1
            return True
        return False

    def expect(type: TokenType, val: Optional[str] = None):
        nonlocal index
        if match(type, val):
            index += 1
            return tokens[index - 1]
        if not has():
            raise Exception("expected {}, got EOL".format(type))
        else:
            raise Exception("expected {}, got {}".format(type, current().type))

    parse_term: Callable[[], None]

    def parse_bin(priority=0):
        if priority >= MAX_PRIORITY:
            return parse_term()

        left = parse_bin(priority + 1)
        ops = ops_by_priority[priority]

        while has() and current().type == TokenType.OPERATION:
            for op in ops:
                if accept(TokenType.OPERATION, op.symbol):
                    right = parse_bin(priority + 1)
                    left = op.perform(left, right)
                    break
            else:
                break

        return left

    def parse_expr():
        return parse_bin()

    def parse_term():
        token = current()

        if token.type == TokenType.NUMBER:
            return expect(TokenType.NUMBER).num
        elif accept(TokenType.PARENTHESIS, "("):
            val = parse_expr()
            expect(TokenType.PARENTHESIS, ")")
            return val
        else:
            raise Exception("expected term, got {}".format(token.type))

    return parse_expr()


if __name__ == "__main__":
    while True:
        # inp = input("> ")
        inp = "2 + 3 * 4"
        try:
            tok = tokenize(inp)
            res = parse(tok)
            print(res)
        except Exception as e:
            # print(e)
            pass
        print()
        break
