# coding: utf-8
import ast
from typing import Iterable

from dataclasses import dataclass, field

from transpiler.phases.typing import FunctionType
from transpiler.phases.typing.scope import Scope
from transpiler.phases.emit_cpp import CoroutineMode, FunctionEmissionKind, NodeVisitor
from transpiler.phases.emit_cpp.block import BlockVisitor
from transpiler.phases.emit_cpp.class_ import ClassVisitor
from transpiler.phases.emit_cpp.function import FunctionVisitor
from transpiler.utils import compare_ast, highlight

IGNORED_IMPORTS = {"typon", "typing", "__future__", "dataclasses", "enum"}

# noinspection PyPep8Naming
@dataclass
class ModuleVisitor(BlockVisitor):
    includes: list[str] = field(default_factory=list)
    def visit_Import(self, node: ast.Import) -> Iterable[str]:
        TB = f"emitting C++ code for {highlight(node)}"
        for alias in node.names:
            concrete = self.fix_name(alias.asname or alias.name)
            if alias.module_obj.is_python:
                yield f"namespace py_{concrete} {{"
                yield f"struct {concrete}_t {{"

                for name, obj in alias.module_obj.members.items():
                    if obj.python_func_used:
                        yield from self.emit_python_func(alias.name, name, name, obj)

                yield "} all;"
                yield f"auto& get_all() {{ return all; }}"
                yield "}"
                yield f'auto& {concrete} = py_{concrete}::get_all();'
            elif alias.name in IGNORED_IMPORTS:
                yield ""
            else:
                yield from self.import_module(alias.name)
                yield f'auto& {concrete} = py_{alias.name}::get_all();'

    def import_module(self, name: str) -> Iterable[str]:
        self.includes.append(f'#include <python/{name}.hpp>')
        yield ""

    def emit_python_func(self, mod: str, name: str, alias: str, fty: FunctionType) -> Iterable[str]:
        TB = f"emitting C++ code for Python function {highlight(f'{mod}.{name}')}"

        yield "struct {"
        yield f"auto operator()("

        for i, argty in enumerate(fty.parameters):
            if i != 0:
                yield ", "
            yield "lvalue_or_rvalue<"
            yield from self.visit(argty)
            yield f"> arg{i}"

        yield ") {"
        yield "py::scoped_interpreter guard{};"
        yield f"return py::module_::import(\"{mod}\").attr(\"{name}\")("
        for i, argty in enumerate(fty.parameters):
            if i != 0:
                yield ", "
            yield f"*arg{i}"
        yield ").cast<"
        yield from self.visit(fty.return_type)
        yield ">();"
        yield "}"
        yield f"}} {alias};"

    def visit_ImportFrom(self, node: ast.ImportFrom) -> Iterable[str]:
        if node.module in IGNORED_IMPORTS:
            yield ""
        elif node.module_obj.is_python:
            for alias in node.names:
                fty = alias.item_obj
                assert isinstance(fty, FunctionType)

                yield from self.emit_python_func(node.module, alias.name, alias.asname or alias.name, fty)
        else:
            yield from self.import_module(node.module)
            for alias in node.names:
                yield f"auto& {alias.asname or alias.name} = py_{node.module}::get_all().{alias.name};"

    def visit_Expr(self, node: ast.Expr) -> Iterable[str]:
        if isinstance(node.value, ast.Str):
            if "\n" in node.value.s:
                yield f"/*{node.value.s}*/"
            else:
                yield f"//{node.value.s}"
        else:
            raise NotImplementedError(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> Iterable[str]:
        yield from ClassVisitor().visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Iterable[str]:
        yield from super().visit_free_func(node, FunctionEmissionKind.DECLARATION)


@dataclass
class ModuleVisitor2(NodeVisitor):
    scope: Scope
    def visit_FunctionDef(self, node: ast.FunctionDef) -> Iterable[str]:
        yield from BlockVisitor(self.scope).visit_free_func(node, FunctionEmissionKind.DEFINITION)

    def visit_AST(self, node: ast.AST) -> Iterable[str]:
        yield ""
        pass
